from __future__ import annotations

from typing import TYPE_CHECKING

from . import matchers

if TYPE_CHECKING:
    from .invocation import StubbedInvocation


def invocations_are_sameish(
    left: StubbedInvocation,
    right: StubbedInvocation,
) -> bool:
    """Structural signature-compatibility checks for continuation bookkeeping.

    Intentionally avoids executing user-provided matcher predicates
    (e.g. `arg_that(...)) while comparing stub signatures.
    """

    return (
        _params_are_sameish(left.params, right.params)
        and _named_params_are_sameish(
            left.named_params,
            right.named_params,
        )
    )


def invocations_have_distinct_captors(
    left: StubbedInvocation,
    right: StubbedInvocation,
) -> bool:
    """Return True when equivalent selectors bind different captor instances."""

    for left_value, right_value in zip(left.params, right.params):
        if _values_bind_distinct_captors(left_value, right_value):
            return True

    for key in set(left.named_params) & set(right.named_params):
        if _values_bind_distinct_captors(
            left.named_params[key],
            right.named_params[key],
        ):
            return True

    return False


def _params_are_sameish(left: tuple, right: tuple) -> bool:
    if len(left) != len(right):
        return False

    return all(
        _values_are_sameish(left_value, right_value)
        for left_value, right_value in zip(left, right)
    )


def _named_params_are_sameish(left: dict, right: dict) -> bool:
    if set(left) != set(right):
        return False

    return all(
        _values_are_sameish(left[key], right[key])
        for key in left
    )


def _values_are_sameish(left: object, right: object) -> bool:
    if left is right:
        return True

    if left is Ellipsis or right is Ellipsis:
        return left is right

    if matchers.is_call_captor(left) and matchers.is_call_captor(right):
        return True

    if matchers.is_call_captor(left) or matchers.is_call_captor(right):
        return False

    if (
        matchers.is_captor_args_sentinel(left)
        and matchers.is_captor_args_sentinel(right)
    ):
        return _values_are_sameish(left.captor.matcher, right.captor.matcher)

    if (
        matchers.is_captor_kwargs_sentinel(left)
        and matchers.is_captor_kwargs_sentinel(right)
    ):
        return _values_are_sameish(left.captor.matcher, right.captor.matcher)

    if (
        matchers.is_captor_args_sentinel(left)
        or matchers.is_captor_args_sentinel(right)
        or matchers.is_captor_kwargs_sentinel(left)
        or matchers.is_captor_kwargs_sentinel(right)
    ):
        return False

    if isinstance(left, matchers.Matcher) and isinstance(right, matchers.Matcher):
        return _matchers_are_sameish(left, right)

    if isinstance(left, matchers.Matcher) or isinstance(right, matchers.Matcher):
        return False

    return _equals_or_identity(left, right)


def _matchers_are_sameish(  # noqa: C901
    left: matchers.Matcher,
    right: matchers.Matcher,
) -> bool:
    if left is right:
        return True

    if type(left) is not type(right):
        return False

    if isinstance(left, matchers.Any) and isinstance(right, matchers.Any):
        return _equals_or_identity(left.wanted_type, right.wanted_type)

    if (
        isinstance(left, matchers.ValueMatcher)
        and isinstance(right, matchers.ValueMatcher)
    ):
        return _values_are_sameish(left.value, right.value)

    if (
        isinstance(left, (matchers.And, matchers.Or))
        and isinstance(right, (matchers.And, matchers.Or))
    ):
        return _params_are_sameish(
            tuple(left.matchers),
            tuple(right.matchers),
        )

    if isinstance(left, matchers.Not) and isinstance(right, matchers.Not):
        return _values_are_sameish(left.matcher, right.matcher)

    if isinstance(left, matchers.ArgThat) and isinstance(right, matchers.ArgThat):
        return left.predicate is right.predicate

    if isinstance(left, matchers.Contains) and isinstance(right, matchers.Contains):
        return _values_are_sameish(left.sub, right.sub)

    if isinstance(left, matchers.Matches) and isinstance(right, matchers.Matches):
        return (
            left.regex.pattern == right.regex.pattern
            and left.flags == right.flags
        )

    if (
        isinstance(left, matchers.ArgumentCaptor)
        and isinstance(right, matchers.ArgumentCaptor)
    ):
        return _values_are_sameish(left.matcher, right.matcher)

    return _equals_or_identity(left, right)


def _values_bind_distinct_captors(left: object, right: object) -> bool:
    left_binding = _captor_binding(left)
    right_binding = _captor_binding(right)

    return (
        left_binding is not None
        and right_binding is not None
        and left_binding is not right_binding
    )


def _captor_binding(value: object) -> object | None:
    if matchers.is_call_captor(value):
        return value

    if isinstance(value, matchers.ArgumentCaptor):
        return value

    if matchers.is_captor_args_sentinel(value):
        return value.captor

    if matchers.is_captor_kwargs_sentinel(value):
        return value.captor

    return None


def _equals_or_identity(left: object, right: object) -> bool:
    try:
        return left == right
    except Exception:
        return left is right



# Copyright (c) 2008-2016 Szczepan Faber, Serhiy Oplakanets, Herr Kaste
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Argument matchers for stubbing and verifications.

In general the call signature you specify when stubbing or verifying in mockito
is as concrete as possible: it consists of values only::

    when(os.path).exists('/foo/bar.txt').thenReturn(True)

This is for a reason. In controlled test environments, for the scope of a
single test, you should usually know exactly how you use a function, and what
you expect its outcome to be. In mockito usually (in `strict` mode) all
invocations you did not specify upfront will throw at call time.

If you reason about your code, the above `when` tirade turns - for the time
of the test - the specific stubbed function into a constant.

You can use so called argument matchers below if you can't or don't
want to specify a single concrete value for an argument, but a type or class of
possible values. E.g.::

    when(os.path).exists(...).thenReturn(True)
    when(os.path).exists(ANY).thenReturn(True)
    when(os.path).exists(ANY(str)).thenReturn(True)

    when(requests).get(ANY(str), **kwargs)
    when(requests).get('https://example.com', ...)

    when(math).sqrt(not_(_or(ANY(float), ANY(int)))).thenRaise(TypeError)

Now what you get each time is a function that up to a degree takes various
arguments and responds with the same outcome each time. Now that's a weird
thing. So use the matchers for a reason, they're powerful.

The one usage you should not care about is a loose signature when using
:func:`verify`. Since mockito will throw for unexpected calls, a very loose
`verify` should be ok::

    verify(requests, times=1).get(...)


"""

from abc import ABC, abstractmethod
import functools
import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    try:
        from typing import TypeGuard
    except ImportError:
        from typing_extensions import TypeGuard

builtin_any = any

__all__ = [
    'and_', 'or_', 'not_',
    'eq', 'neq',
    'lt', 'lte',
    'gt', 'gte',
    'any', 'any_', 'ANY',
    'arg_that',
    'contains',
    'matches',
    'captor',
    'call_captor',
    'times',
    'args', 'ARGS',
    'kwargs', 'KWARGS'
]


class _ArgsSentinel(object):
    def __repr__(self):
        return '*args'


ARGS_SENTINEL = _ArgsSentinel()
ARGS = args = [ARGS_SENTINEL]
# ARGS.__doc__ = """Matches multiple positional arguments.

# Note: `args` must match at least one argument.

# Example::

#     when(manager).add_tasks(1, 2, *args)

# """

KWARGS_SENTINEL = '**'
KWARGS = kwargs = {KWARGS_SENTINEL: '_'}
# KWARGS.__doc__ = """Matches multiple keyword arguments.

# Note that `kwargs` must match at least one remaining keyword argument.

# Example::

#     when(requests).get('http://myapi/', **KWARGS)

# """


class MatcherError(RuntimeError):
    '''Indicates generic runtime error raised by mockito-python matchers
    '''
    pass


class Matcher:
    def matches(self, arg):
        pass

class Capturing(ABC):
    @abstractmethod
    def capture_value(self, value):
        pass


class Any(Matcher):
    def __init__(self, wanted_type=None):
        self.wanted_type = wanted_type

    def matches(self, arg):
        if self.wanted_type:
            return isinstance(arg, self.wanted_type)
        else:
            return True

    def __repr__(self):
        return "<Any: %s>" % _any_wanted_type_label(self.wanted_type)


def _any_wanted_type_label(wanted_type):
    if isinstance(wanted_type, type):
        return _type_label(wanted_type)

    if (
        isinstance(wanted_type, tuple)
        and all(isinstance(t, type) for t in wanted_type)
    ):
        items = [_type_label(t) for t in wanted_type]
        if len(items) == 1:
            return '(%s,)' % items[0]
        return '(%s)' % ', '.join(items)

    return _safe_repr(wanted_type)


def _type_label(type_):
    module = _safe_getattr(type_, '__module__')
    qualname = _safe_getattr(type_, '__qualname__') or _safe_getattr(type_, '__name__')
    if qualname is None:
        return _safe_repr(type_)

    if module is None or module == 'builtins':
        return qualname

    return '%s.%s' % (module, qualname)


class ValueMatcher(Matcher):
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return "<%s: %s>" % (
            self.__class__.__name__,
            _safe_repr(self.value),
        )


class Eq(ValueMatcher):
    def matches(self, arg):
        return arg == self.value


class Neq(ValueMatcher):
    def matches(self, arg):
        return arg != self.value


class Lt(ValueMatcher):
    def matches(self, arg):
        return arg < self.value


class Lte(ValueMatcher):
    def matches(self, arg):
        return arg <= self.value


class Gt(ValueMatcher):
    def matches(self, arg):
        return arg > self.value


class Gte(ValueMatcher):
    def matches(self, arg):
        return arg >= self.value


class And(Matcher):
    def __init__(self, matchers):
        self.matchers = [
            matcher if isinstance(matcher, Matcher) else Eq(matcher)
            for matcher in matchers]

    def matches(self, arg):
        return all(matcher.matches(arg) for matcher in self.matchers)

    def __repr__(self):
        return "<And: %s>" % self.matchers


class Or(Matcher):
    def __init__(self, matchers):
        self.matchers = [
            matcher if isinstance(matcher, Matcher) else Eq(matcher)
            for matcher in matchers]

    def matches(self, arg):
        return builtin_any([matcher.matches(arg) for matcher in self.matchers])

    def __repr__(self):
        return "<Or: %s>" % self.matchers


class Not(Matcher):
    def __init__(self, matcher):
        self.matcher = matcher if isinstance(matcher, Matcher) else Eq(matcher)

    def matches(self, arg):
        return not self.matcher.matches(arg)

    def __repr__(self):
        return "<Not: %s>" % self.matcher


class ArgThat(Matcher):
    def __init__(self, predicate):
        self.predicate = predicate

    def matches(self, arg):
        return self.predicate(arg)

    def __repr__(self):
        return "<ArgThat: %s>" % _arg_that_predicate_label(self.predicate)


def _arg_that_predicate_label(predicate):
    try:
        return _arg_that_predicate_label_unchecked(predicate)
    except Exception:
        predicate_class = _safe_getattr(
            _safe_getattr(predicate, '__class__'),
            '__name__',
        )
        if predicate_class is None:
            return 'callable'

        return 'callable %s' % predicate_class


def _arg_that_predicate_label_unchecked(predicate):
    if isinstance(predicate, functools.partial):
        return _arg_that_partial_label(predicate)

    function_line = _line_of_callable(predicate)
    function_name = _safe_getattr(predicate, '__name__')
    if function_name is not None:
        if function_name == '<lambda>':
            return _label_with_line('lambda', function_line)
        return _label_with_line('def %s' % function_name, function_line)

    predicate_class = _safe_getattr(
        _safe_getattr(predicate, '__class__'),
        '__name__',
    )
    if predicate_class is None:
        predicate_class = 'object'

    call = _safe_getattr(predicate, '__call__')
    call_line = _line_of_callable(call)
    return _label_with_line(
        'callable %s.__call__' % predicate_class,
        call_line,
    )


def _arg_that_partial_label(predicate):
    partial_func = _safe_getattr(predicate, 'func')
    partial_name = _safe_getattr(partial_func, '__name__')

    if partial_name is not None:
        return 'partial %s' % partial_name

    return 'partial'


def _line_of_callable(value):
    if value is None:
        return None

    func = _safe_getattr(value, '__func__', value)
    code = _safe_getattr(func, '__code__')
    if code is None:
        return None

    return _safe_getattr(code, 'co_firstlineno')


def _safe_getattr(value, name, default=None):
    try:
        return getattr(value, name)
    except Exception:
        return default


def _safe_repr(value):
    try:
        return repr(value)
    except Exception:
        try:
            return object.__repr__(value)
        except Exception:
            return '<unrepresentable>'


def _label_with_line(label, line_number):
    if line_number is None:
        return label

    return '%s at line %s' % (label, line_number)


class Contains(Matcher):
    def __init__(self, sub):
        self.sub = sub

    def matches(self, arg):
        if not hasattr(arg, 'find'):
            return
        return self.sub and len(self.sub) > 0 and arg.find(self.sub) > -1

    def __repr__(self):
        return "<Contains: %s>" % _safe_repr(self.sub)


class Matches(Matcher):
    def __init__(self, regex, flags=0):
        self.regex = re.compile(regex, flags)
        self.flags = _explicit_regex_flags(regex, flags)

    def matches(self, arg):
        if not isinstance(arg, str):
            return
        return self.regex.match(arg) is not None

    def __repr__(self):
        if self.flags:
            return "<Matches: %r flags=%d>" % (self.regex.pattern, self.flags)
        else:
            return "<Matches: %r>" % self.regex.pattern


def _explicit_regex_flags(regex, flags):
    if flags:
        return flags

    compiled_flags = _safe_getattr(regex, 'flags')
    pattern = _safe_getattr(regex, 'pattern')
    if compiled_flags is None or pattern is None:
        return 0

    try:
        baseline_flags = re.compile(pattern).flags
    except Exception:
        return compiled_flags

    return compiled_flags & ~baseline_flags


class ArgumentCaptor(Matcher, Capturing):
    def __init__(self, matcher=None):
        self.matcher = matcher or Any()
        self.all_values = []

    def matches(self, arg):
        result = self.matcher.matches(arg)
        if not result:
            return
        return True

    def __iter__(self):
        yield CaptorArgsSentinel(self)

    def keys(self):
        return [KWARGS_SENTINEL]

    def __getitem__(self, key):
        if key is not KWARGS_SENTINEL and key != KWARGS_SENTINEL:
            raise KeyError(key)
        return CaptorKwargsSentinel(self)

    @property
    def value(self):
        if not self.all_values:
            raise MatcherError("No argument value was captured!")
        return self.all_values[-1]

    def capture_value(self, value):
        self.all_values.append(value)

    def __repr__(self):
        return "<ArgumentCaptor: matcher=%s values=%s>" % (
            repr(self.matcher), self.all_values,
        )


class CallCaptor:
    def __init__(self):
        self.all_values = []

    @property
    def value(self):
        if not self.all_values:
            raise MatcherError("No call value was captured!")
        return self.all_values[-1]

    def capture_call(self, args, kwargs):
        self.all_values.append((tuple(args), dict(kwargs)))

    def __repr__(self):
        return "<CallCaptor: values=%s>" % self.all_values


class CaptorArgsSentinel:
    def __init__(self, captor):
        self.captor = captor

    def matches(self, args):
        return all(self.captor.matches(arg) for arg in args)

    def capture_value(self, value):
        self.captor.capture_value(value)

    def __repr__(self):
        return "<CaptorArgsSentinel: %r>" % self.captor


class CaptorKwargsSentinel:
    def __init__(self, captor):
        self.captor = captor

    def matches(self, kwargs):
        return all(self.captor.matches(value) for value in kwargs.values())

    def capture_value(self, value):
        self.captor.capture_value(value)

    def __repr__(self):
        return "<CaptorKwargsSentinel: %r>" % self.captor


def is_call_captor(value: object) -> 'TypeGuard[CallCaptor]':
    return isinstance(value, CallCaptor)


def is_captor_args_sentinel(value: object) -> 'TypeGuard[CaptorArgsSentinel]':
    return isinstance(value, CaptorArgsSentinel)


def is_captor_kwargs_sentinel(value: object) -> 'TypeGuard[CaptorKwargsSentinel]':
    return isinstance(value, CaptorKwargsSentinel)


def is_args_sentinel(value):
    return value is ARGS_SENTINEL or is_captor_args_sentinel(value)


def any(wanted_type=None):
    """Matches against type of argument (`isinstance`).

    If you want to match *any* type, use either `ANY` or `ANY()`.

    Examples::

        when(mock).foo(any).thenReturn(1)
        verify(mock).foo(any(int))

    """
    return Any(wanted_type)


ANY = any_ = any


def eq(value):
    """Matches particular value (`==`)"""
    return Eq(value)


def neq(value):
    """Matches any but given value (`!=`)"""
    return Neq(value)


def lt(value):
    """Matches any value that is less than given value (`<`)"""
    return Lt(value)


def lte(value):
    """Matches any value that is less than or equal to given value (`<=`)"""
    return Lte(value)


def gt(value):
    """Matches any value that is greater than given value (`>`)"""
    return Gt(value)


def gte(value):
    """Matches any value that is greater than or equal to given value (`>=`)"""
    return Gte(value)


def and_(*matchers):
    """Matches if all given matchers match

    Example::

        when(mock).foo(and_(ANY(str), contains('foo')))

    """
    return And(matchers)


def or_(*matchers):
    """Matches if any given matcher match

    Example::

        when(mock).foo(or_(ANY(int), ANY(float)))

    """
    return Or(matchers)


def not_(matcher):
    """Matches if given matcher does not match

    Example::

        when(mock).foo(not_(ANY(str))).thenRaise(TypeError)

    """
    return Not(matcher)


def arg_that(predicate):
    """Matches any argument for which predicate returns True

    Example::

        verify(mock).foo(arg_that(lambda arg: arg > 3 and arg < 7))

    """
    return ArgThat(predicate)


def contains(sub):
    """Matches any string containing given substring

    Example::

        mock.foo([120, 121, 122, 123])
        verify(mock).foo(contains(123))

    """
    return Contains(sub)


def matches(regex, flags=0):
    """Matches any string that matches given regex"""
    return Matches(regex, flags)


def captor(matcher=None):
    """Returns argument captor that captures values for further assertions

    Example::

        arg = captor()
        mock.do_something(123)
        mock.do_something(456)
        verify(mock).do_something(arg)
        assert arg.value == 456
        assert arg.all_values == [123, 456]

    You can restrict what the captor captures using the other matchers
    shown herein::

        arg = captor(any(str))
        arg = captor(contains("foo"))

    captor can also be used to capture rest arguments::

        args = captor()
        kwargs = captor()
        when(mock).do(*args, **kwargs)
    """
    return ArgumentCaptor(matcher)


def call_captor():
    """Returns a call captor that captures ``(args, kwargs)`` tuples.

    Example::

        call = call_captor()
        when(mock).do(call).thenReturn("ok")
        mock.do(1, 2, x=3)
        assert call.value == ((1, 2), {"x": 3})

    """
    return CallCaptor()


def times(count):
    return count



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import MutableMapping
from contextlib import contextmanager
from dataclasses import dataclass
import inspect

from .utils import MISSING_ATTRIBUTE, get_original_attribute


@dataclass
class _RestoreInformation:
    obj: object
    attr_name: str
    original_value: object
    use_set_on_restore: bool


@contextmanager
def restore_patch_contextmanager(patch: Patch, yield_value: object = None):
    try:
        yield yield_value
    finally:
        patch.restore_and_unregister()


class Patcher:
    def __init__(self) -> None:
        self._patches: list[Patch] = []
        self._restore_infos: list[_RestoreInformation] = []

    def patch_attribute(
        self,
        obj: object,
        attr_name: str,
        replacement: object,
        *,
        allow_unstub_by_replacement: bool,
    ) -> _AttrPatch:
        attr_patch = _AttrPatch(
            registry=self,
            obj=obj,
            attr_name=attr_name,
            replacement=replacement,
            allow_unstub_by_replacement=allow_unstub_by_replacement,
        )
        attr_patch.apply()
        self._register_patch(attr_patch)
        return attr_patch

    def patch_dictionary(
        self,
        target: MutableMapping[object, object],
        updates: dict[object, object],
        *,
        clear: bool = False,
        remove: tuple[object, ...] = (),
    ) -> _DictPatch:
        dict_patch = _DictPatch(
            registry=self,
            target=target,
            updates=updates,
            clear=clear,
            remove=remove,
        )
        dict_patch.apply()
        self._register_patch(dict_patch)
        return dict_patch

    def unstub_matching(self, obj: object) -> bool:
        matching = [
            patch for patch in self._patches
            if patch.matches_unstub_target(obj)
        ]
        for patch in reversed(matching):
            patch.restore_and_unregister()

        return bool(matching)

    def unstub_attribute(self, obj: object, attr_name: str) -> bool:
        matching = [
            patch for patch in self._patches
            if isinstance(patch, _AttrPatch)
            if patch.obj is obj
            if patch.attr_name == attr_name
        ]
        for patch in reversed(matching):
            patch.restore_and_unregister()

        return bool(matching)

    def unstub_all(self) -> None:
        for patch in reversed(self._patches.copy()):
            patch.restore_and_unregister()

    def unregister_patch(self, patch: Patch) -> None:
        try:
            self._patches.remove(patch)
        except ValueError:
            pass

    def _register_patch(self, patch: Patch) -> None:
        self._patches.append(patch)

    @contextmanager
    def capture_restore_information(self, patch: _AttrPatch):
        has_restore_info = self.has_restore_information(patch.obj, patch.attr_name)

        if not has_restore_info:
            restore_info = _capture_restore_information(patch.obj, patch.attr_name)

        try:
            yield
        except Exception:
            raise
        else:
            if not has_restore_info:
                self._restore_infos.append(restore_info)

    def stack_for_attr_patch(self, patch: _AttrPatch) -> list[_AttrPatch]:
        return [
            candidate
            for candidate in reversed(self._patches)
            if isinstance(candidate, _AttrPatch)
            if candidate.active
            if candidate.obj is patch.obj
            if candidate.attr_name == patch.attr_name
        ]

    def has_restore_information(self, obj: object, attr_name: str) -> bool:
        return self.find_restore_information(obj, attr_name) is not None

    def find_restore_information(
        self, obj: object, attr_name: str
    ) -> _RestoreInformation | None:
        for restore_info in self._restore_infos:
            if restore_info.obj is obj and restore_info.attr_name == attr_name:
                return restore_info
        return None

    def remove_restore_information(self, restore_info: _RestoreInformation) -> None:
        try:
            self._restore_infos.remove(restore_info)
        except ValueError:
            pass


def _capture_restore_information(obj: object, attr_name: str) -> _RestoreInformation:
    original_value, use_set_on_restore = get_original_attribute(
        obj, attr_name, default=MISSING_ATTRIBUTE
    )

    if (
        not use_set_on_restore
        and original_value is not MISSING_ATTRIBUTE
        and _has_data_descriptor_on_type(obj, attr_name)
    ):
        use_set_on_restore = True

    return _RestoreInformation(
        obj=obj,
        attr_name=attr_name,
        original_value=original_value,
        use_set_on_restore=use_set_on_restore,
    )


def _restore_original_attribute(restore_info: _RestoreInformation) -> None:
    if restore_info.use_set_on_restore:
        setattr(
            restore_info.obj,
            restore_info.attr_name,
            restore_info.original_value
        )
        return

    try:
        delattr(restore_info.obj, restore_info.attr_name)
    except AttributeError:
        pass


class Patch(ABC):
    def __init__(self, registry: Patcher) -> None:
        self.registry = registry
        self.active = False

    @abstractmethod
    def apply(self) -> None:
        pass

    @abstractmethod
    def restore(self) -> None:
        pass

    @abstractmethod
    def matches_unstub_target(self, obj: object) -> bool:
        pass

    def restore_and_unregister(self) -> None:
        try:
            self.restore()
        finally:
            self.registry.unregister_patch(self)


class _AttrPatch(Patch):
    def __init__(
        self,
        registry: Patcher,
        obj: object,
        attr_name: str,
        replacement: object,
        *,
        allow_unstub_by_replacement: bool,
    ):
        super().__init__(registry)
        self.obj = obj
        self.attr_name = attr_name
        self.replacement = replacement
        self.allow_unstub_by_replacement = allow_unstub_by_replacement

    def apply(self) -> None:
        if self.active:
            return

        with self.registry.capture_restore_information(self):
            setattr(self.obj, self.attr_name, self.replacement)

        self.active = True

    def restore(self) -> None:
        if not self.active:
            return

        stack = self.registry.stack_for_attr_patch(self)
        if not stack or stack[0] is not self:
            return

        if len(stack) > 1:
            setattr(self.obj, self.attr_name, stack[1].replacement)
            return

        restore_info = self.registry.find_restore_information(self.obj, self.attr_name)
        if restore_info:
            _restore_original_attribute(restore_info)
            self.registry.remove_restore_information(restore_info)

        self.active = False

    def matches_unstub_target(self, obj: object) -> bool:
        return self.obj is obj or (
            self.allow_unstub_by_replacement and self.replacement is obj
        )


class _DictPatch(Patch):
    def __init__(
        self,
        registry: Patcher,
        target: MutableMapping[object, object],
        updates: dict[object, object],
        *,
        clear: bool,
        remove: tuple[object, ...],
    ):
        super().__init__(registry)
        self.target = target
        self.updates = updates
        self.clear = clear
        self.remove = remove

        self.original: dict[object, object] = {}

    def apply(self) -> None:
        if self.active:
            return

        self.original = dict(self.target)

        try:
            if self.clear:
                self.target.clear()
            else:
                for key in self.remove:
                    self.target.pop(key, None)

            self.target.update(self.updates)
        except Exception:
            self.target.clear()
            self.target.update(self.original)
            raise

        self.active = True

    def restore(self) -> None:
        if not self.active:
            return

        self.target.clear()
        self.target.update(self.original)

        self.active = False

    def matches_unstub_target(self, obj: object) -> bool:
        return self.target is obj


def _has_data_descriptor_on_type(obj: object, attr_name: str) -> bool:
    if inspect.isclass(obj):
        return False

    try:
        type_attr = inspect.getattr_static(type(obj), attr_name)
    except AttributeError:
        return False

    return hasattr(type_attr, "__set__") or hasattr(type_attr, "__delete__")


patcher = Patcher()



# Copyright (c) 2008-2016 Szczepan Faber, Serhiy Oplakanets, Herr Kaste
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

import operator
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .invocation import MatchingInvocation

__all__ = ['never', 'VerificationError']


class VerificationError(AssertionError):
    '''Indicates error during verification of invocations.

    Raised if verification fails. Error message contains the cause.
    '''
    pass


__tracebackhide__ = operator.methodcaller("errisinstance", VerificationError)


class VerificationMode(ABC):
    @abstractmethod
    def verify(
        self, invocation: MatchingInvocation, actual_count: int
    ) -> None:
        pass


class AtLeast(VerificationMode):
    def __init__(self, wanted_count: int) -> None:
        self.wanted_count = wanted_count

    def verify(
        self, invocation: MatchingInvocation, actual_count: int
    ) -> None:
        if actual_count == 0:
            msg = error_message_for_unmatched_invocation(invocation)
            raise VerificationError(msg)

        if actual_count < self.wanted_count:
            raise VerificationError("\nWanted at least: %i, actual times: %i"
                                    % (self.wanted_count, actual_count))

    def __repr__(self):
        return "<%s wanted=%s>" % (type(self).__name__, self.wanted_count)

class AtMost(VerificationMode):
    def __init__(self, wanted_count: int) -> None:
        self.wanted_count = wanted_count

    def verify(
        self, invocation: MatchingInvocation, actual_count: int
    ) -> None:
        if actual_count > self.wanted_count:
            raise VerificationError("\nWanted at most: %i, actual times: %i"
                                    % (self.wanted_count, actual_count))

    def __repr__(self):
        return "<%s wanted=%s>" % (type(self).__name__, self.wanted_count)

class Between(VerificationMode):
    def __init__(
        self, wanted_from: int, wanted_to: int | float = float('inf')
    ) -> None:
        self.wanted_from = wanted_from
        self.wanted_to = wanted_to

    def verify(
        self, invocation: MatchingInvocation, actual_count: int
    ) -> None:
        if actual_count < self.wanted_from or actual_count > self.wanted_to:
            raise VerificationError(
                "\nWanted between: [%s, %s], actual times: %s"
                % (self.wanted_from, self.wanted_to, actual_count))

    def __repr__(self):
        return "<Between [%s, %s]>" % (self.wanted_from, self.wanted_to)

class Times(VerificationMode):
    def __init__(self, wanted_count: int) -> None:
        self.wanted_count = wanted_count

    def verify(
        self, invocation: MatchingInvocation, actual_count: int
    ) -> None:
        if actual_count == self.wanted_count:
            return

        if actual_count == 0:
            msg = error_message_for_unmatched_invocation(invocation)
            raise VerificationError(msg)

        if self.wanted_count == 0:
            raise VerificationError(
                "\nUnwanted invocation of %s, times: %i"
                % (invocation, actual_count))
        else:
            raise VerificationError("\nWanted times: %i, actual times: %i"
                                    % (self.wanted_count, actual_count))

    def __repr__(self):
        return "<%s wanted=%s>" % (type(self).__name__, self.wanted_count)


def error_message_for_unmatched_invocation(
    invocation: MatchingInvocation
) -> str:
    invocations = (
        [
            invoc
            for invoc in invocation.mock.invocations
            if invoc.method_name == invocation.method_name
        ]
        or invocation.mock.invocations
    )
    wanted_section = (
        f"\nWanted but not invoked:\n\n    {invocation}\n"
    )
    if invocations:
        content = "\n    ".join(map(str, invocations))
        instead_section = f"\nInstead got:\n\n    {content}\n"
    elif (
        len(invocation.mock.stubbed_invocations) > 1
        or len(invocation.mock.stubbed_invocations) == 1
        and str(invocation.mock.stubbed_invocations[0]) != str(invocation)
    ):
        content = "\n    ".join(sorted(
            map(str, invocation.mock.stubbed_invocations)
        ))
        instead_section = f"\nStubbed are:\n\n    {content}\n"
    else:
        instead_section = ""

    return "%s%s\n" % (wanted_section, instead_section)


class InOrder(VerificationMode):
    '''Verifies invocations in order.

    Verifies if invocation was in expected order, and if yes -- degrades to
    original Verifier (AtLeast, Times, Between, ...).
    '''

    def __init__(self, original_verification: VerificationMode) -> None:
        '''

        @param original_verification: Original verification to degrade to if
                                      order of invocation was ok.
        '''
        self.original_verification = original_verification

    def verify(
        self, wanted_invocation: MatchingInvocation, count: int
    ) -> None:
        for invocation in wanted_invocation.mock.invocations:
            if not invocation.verified_inorder:
                if not wanted_invocation.matches(invocation):
                    raise VerificationError(
                        '\nWanted %s to be invoked,'
                        '\ngot    %s instead.' %
                        (wanted_invocation, invocation))
                invocation.verified_inorder = True
                break
        # proceed with original verification
        self.original_verification.verify(wanted_invocation, count)


never = 0



# Copyright (c) 2008-2016 Szczepan Faber, Serhiy Oplakanets, Herr Kaste
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from __future__ import annotations

import ast
import inspect
from collections import deque
from functools import partial
from typing import Deque, TYPE_CHECKING

from .verification import VerificationError
from .invocation import RealInvocation, VerifiableInvocation
from .mockito import ArgumentError, verify as verify_main
from .mock_registry import mock_registry
from .utils import deprecated

if TYPE_CHECKING:
    from .mocking import Mock


@deprecated(
    "'inorder.verify' is deprecated. "
    "Use 'InOrder(...).verify(...)' instead."
)
def verify(object, *args, **kwargs):
    kwargs['inorder'] = True
    return verify_main(object, *args, **kwargs)


class InOrder:
    """Verify interactions in strict order across one or multiple objects.

    This is the preferred API for order-sensitive verification. `InOrder`
    keeps one global queue of recorded invocations for all observed objects,
    then verifies against that queue from left to right.

    Basic usage::

        from mockito import mock, InOrder

        cat = mock()
        dog = mock()

        in_order = InOrder(cat, dog)
        cat.meow()
        dog.bark()

        in_order.verify(cat).meow()
        in_order.verify(dog).bark()

    `InOrder.verify` uses the same fluent style and verification arguments as
    :func:`verify`, including ``times``, ``atleast``, ``atmost`` and
    ``between``::

        in_order.verify(cat, times=2).meow(...)
        in_order.verify(dog, between=(1, 3)).bark()

    Zero-lower-bound verifications (e.g. ``times=0`` or ``between=(0, 2)``)
    may pass without consuming the next queued invocation.

    `InOrder` can be used as a context manager::

        with InOrder(cat, dog) as in_order:
            cat.meow()
            dog.bark()
            in_order.verify(cat).meow()
            in_order.verify(dog).bark()

    Only the objects passed to `InOrder` are observed. Calls on other objects
    are ignored. Registration is dynamic: objects that are stubbed after
    `InOrder(...)` construction are still captured while the instance is active.

    See related :func:`verify`, :func:`expect`,
    :func:`ensureNoUnverifiedInteractions`, and
    :func:`verifyStubbedInvocationsAreUsed`.

    """

    def __init__(self, *objects: object) -> None:
        self._core = InOrderImpl(*objects)

    def verify(
        self,
        obj: object,
        times=None,
        atleast=None,
        atmost=None,
        between=None,
    ):
        return self._core.verify(
            obj,
            times=times,
            atleast=atleast,
            atmost=atmost,
            between=between,
        )

    def __enter__(self):
        self._core.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._core.__exit__(exc_type, exc_val, exc_tb)


class InOrderImpl:
    """Internal implementation behind the public `InOrder` facade."""


    def __init__(self, *objects: object):
        objects_: list[object] = []
        for obj in objects:
            if any(observed is obj for observed in objects_):
                raise ValueError(f"{obj} is provided more than once")
            objects_.append(obj)
        self._objects = objects_
        self._object_labels = self._guess_object_labels_from_callsite(
            len(objects_)
        )
        self._active = True
        self._observer_registered = False
        self.ordered_invocations: Deque[RealInvocation] = deque()

        self._register_observer()
        self._attach_all()

    def _is_observed(self, obj: object) -> bool:
        return any(observed is obj for observed in self._objects)

    def _register_observer(self) -> None:
        if not self._observer_registered:
            mock_registry.add_register_observer(self._on_mock_registered)
            self._observer_registered = True

    def _unregister_observer(self) -> None:
        if self._observer_registered:
            mock_registry.remove_register_observer(self._on_mock_registered)
            self._observer_registered = False

    def _attach_all(self):
        for obj in self._objects:
            if m := mock_registry.mock_for(obj):
                m.attach(self)

    def _detach_all(self) -> None:
        for obj in self._objects:
            if m := mock_registry.mock_for(obj):
                m.detach(self)

    def _on_mock_registered(self, obj: object, mock: Mock) -> None:
        if self._active and self._is_observed(obj):
            mock.attach(self)

    def update(self, invocation: RealInvocation) -> None:
        self.ordered_invocations.append(invocation)

    def verify(
        self,
        obj: object,
        times=None,
        atleast=None,
        atmost=None,
        between=None,
    ):
        """
        Central method of InOrder class.
        Use this method to verify the calling order of observed mocks.
        :param obj: obj to verify the ordered invocation

        """
        expected_mock = mock_registry.mock_for(obj)
        if expected_mock is None:
            raise ArgumentError(
                f"\n{obj} is not setup with any stubbings or expectations."
            )

        if not self._is_observed(obj):
            raise ArgumentError(
                f"\n{obj} is not part of that InOrder."
            )

        return verify_main(
            obj=obj,
            times=times,
            atleast=atleast,
            atmost=atmost,
            between=between,
            _factory=partial(InOrderVerifiableInvocation, inorder=self),
        )

    def close(self) -> None:
        self._active = False
        self._detach_all()
        self._unregister_observer()

    def __enter__(self):
        self._active = True
        self._register_observer()
        self._attach_all()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _callsite_label_for_obj(self, obj: object) -> str | None:
        for observed, label in zip(self._objects, self._object_labels):
            if observed is obj:
                return label
        return None

    def _label_for_obj(self, obj: object) -> str:
        if label := self._callsite_label_for_obj(obj):
            return label
        return str(obj)

    def _label_for_mock(self, mock: Mock) -> str:
        obj = mock_registry.obj_for(mock)
        if obj is None:
            return str(mock)
        return self._label_for_obj(obj)

    def _format_invocation(self, mock: Mock, invocation: object) -> str:
        if len(self._objects) == 1:
            return str(invocation)
        return f"{self._label_for_mock(mock)}.{invocation}"

    def _guess_object_labels_from_callsite(self, count: int) -> list[str | None]:
        if count == 0:
            return []

        # For a single observed object, object qualification adds little value
        # to mismatch messages (there is no cross-object ambiguity anyway).
        if count == 1:
            return [None]

        frame = inspect.currentframe()
        # Start at the direct caller and then walk out of our own module.
        # This is important because public `InOrder` delegates to `InOrderImpl`,
        # so the first stack frames are internal wrappers.
        caller_frame = frame.f_back if frame else None
        while caller_frame and caller_frame.f_code.co_filename == __file__:
            caller_frame = caller_frame.f_back

        if caller_frame is None:
            return [None] * count

        try:
            source_lines, start_lineno = inspect.getsourcelines(
                caller_frame.f_code
            )
            source = ''.join(source_lines)
            call_lineno = caller_frame.f_lineno - start_lineno + 1
            tree = ast.parse(source)
        except (OSError, TypeError, SyntaxError):
            return [None] * count
        finally:
            del frame
            del caller_frame

        call = self._find_inorder_call_for_lineno(tree, call_lineno)
        if call is None:
            return [None] * count

        labels: list[str | None] = []
        for arg in call.args[:count]:
            segment = ast.get_source_segment(source, arg)
            labels.append(segment.strip() if segment else None)

        if len(labels) < count:
            labels.extend([None] * (count - len(labels)))

        return labels

    def _find_inorder_call_for_lineno(
        self,
        tree: ast.AST,
        lineno: int,
    ) -> ast.Call | None:
        candidate: ast.Call | None = None
        for node in ast.walk(tree):
            if not isinstance(node, ast.Call):
                continue
            if not self._is_inorder_constructor_call(node):
                continue

            start = node.lineno
            end = getattr(node, 'end_lineno', node.lineno)
            if start <= lineno <= end:
                # Ambiguous callsite (e.g. nested / same-line InOrder calls):
                # bail out.
                if candidate is not None:
                    return None

                candidate = node

        return candidate

    def _is_inorder_constructor_call(self, node: ast.Call) -> bool:
        if isinstance(node.func, ast.Name):
            return node.func.id == 'InOrder'

        if isinstance(node.func, ast.Attribute):
            return node.func.attr == 'InOrder'

        return False


class InOrderVerifiableInvocation(VerifiableInvocation):
    def __init__(self, mock, method_name, verification, inorder: InOrderImpl):
        super().__init__(mock, method_name, verification)
        self._inorder = inorder

    def __call__(self, *params, **named_params) -> None:  # noqa: C901
        self._remember_params(params, named_params)

        ordered = self._inorder.ordered_invocations

        if not ordered:
            if self.handle_zero_matches_if_allowed():
                return
            raise VerificationError(
                "\nThere are no recorded invocations."
            )

        # Find first invocation in global order that hasn't been used
        # for "in-order" verification yet.
        try:
            start_idx, next_invocation = next(
                (i, inv)
                for i, inv in enumerate(ordered)
                if not inv.verified_inorder
            )
        except StopIteration:
            if self.handle_zero_matches_if_allowed():
                return
            raise VerificationError(
                "\nThere are no more recorded invocations."
            )

        called_mock = next_invocation.mock
        if called_mock is not self.mock:
            if self.handle_zero_matches_if_allowed():
                return
            if mock_registry.obj_for(called_mock) is None:
                raise RuntimeError(
                    f"{called_mock} is not in the registry (anymore)."
                )

            expected_obj = mock_registry.obj_for(self.mock)
            expected_has_callsite_label = (
                expected_obj is not None
                and self._inorder._callsite_label_for_obj(expected_obj)
            )
            got = self._inorder._format_invocation(called_mock, next_invocation)

            if expected_has_callsite_label:
                wanted = self._inorder._format_invocation(self.mock, self)
                raise VerificationError(
                    "\nWanted %s to be invoked,"
                    "\ngot    %s instead." % (wanted, got)
                )

            raise VerificationError(
                f"\nWanted a call from {expected_obj},"
                f"\ngot    {got} instead."
            )

        matched_invocations: list[RealInvocation] = []

        # Walk the contiguous block of this mock in the global queue.
        for inv in list(ordered)[start_idx:]:
            if inv.verified_inorder:
                continue
            if inv.mock is not self.mock:
                break

            if not self.matches(inv):
                if not matched_invocations:
                    if self.handle_zero_matches_if_allowed():
                        return
                    raise VerificationError(
                        "\nWanted %s to be invoked,"
                        "\ngot    %s instead." % (self, inv)
                    )
                break

            self.capture_arguments(inv)
            matched_invocations.append(inv)

        self.verification.verify(self, len(matched_invocations))

        for inv in matched_invocations:
            inv.verified = True
            inv.verified_inorder = True

        self.maybe_check_stubs_as_used()

    def handle_zero_matches_if_allowed(self) -> bool:
        if not self.verification_allows_zero_matches:
            return False

        self.verification.verify(self, 0)
        self.maybe_check_stubs_as_used()
        return True
